LANGUAGES_DATE_STRING = {
    "bg": "d MMM yyyy",  # Bulgarian
    # "bn": "d MMM, yyyy",  # Bengali  # disabled for now
    "cs": "d. MMM yyyy",  # Czech
    "de": "d. MMM yyyy",  # German
    "en": "MMM. d, yyyy",  # English
    "es": "d MMM yyyy",  # Spanish
    "eu": "yyyy MMM d",  # Basque
    "fr": "d MMM yyyy",  # French
    "gl": "d MMM yyyy",  # Galician
    "hr": "d. MMM yyyy.",  # Croatian
    "it": "d MMM yyyy",  # Italian
    "nb": "d. MMM yyyy",  # Norwegian Bokmål
    "nl": "d MMM. yyyy",  # Dutch
    "pl": "d MMM yyyy",  # Polish
    "pt": "d MMM yyyy",  # Portuguese
    "ru": "d MMM yyyy",  # Russian
    "zh_Hans": "y年M月d日",  # Chinese (Simplified Han script)
}

SUPPORTED_LANGUAGES = list(LANGUAGES_DATE_STRING.keys())
