{% extends "layout.html" %}
{% block title %}{{ _('Confirm your account') }}{% endblock %}
{% block preheader %}{{ _('Use this link to confirm your account.') }}{% endblock %}
{% block content %}<p>{{ _('You have created an account on FitTrackee.') }} {{ _('Use the button below to confirm your address email.') }}</p>
                        <table class="body-action" align="center" width="100%" cellpadding="0" cellspacing="0" role="presentation">
                          <tr>
                            <td align="center">
                              <table width="100%" border="0" cellspacing="0" cellpadding="0" role="presentation">
                                <tr>
                                  <td align="center">
                                    <a href="{{account_confirmation_url}}" class="f-fallback button button--green" target="_blank">{{ _('Verify your email') }}</a>
                                  </td>
                                </tr>
                              </table>
                            </td>
                          </tr>
                        </table>{% endblock %}
{% block not_initiated %}{{ _("If this account creation wasn't initiated by you, please ignore this email.") }}{% endblock %}
{% block url_to_paste %}<table class="body-sub" role="presentation">
                          <tr>
                            <td>
                              <p class="f-fallback sub">{{ _("If you're having trouble with the button above, copy and paste the URL below into your web browser.") }}</p>
                              <p class="f-fallback sub">{{account_confirmation_url}}</p>
                            </td>
                          </tr>
                        </table>{% endblock %}
